"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/durus/__init__.py $
$Id: __init__.py 29185 2007-01-11 20:51:49Z dbinger $
Copyright (c) Corporation for National Research Initiatives 2007. All Rights Reserved.
"""
__version__ = '3.7'
